(* Abstract Syntax Tree and functions for printing it *)

type op =
    Add
  | And
  | Div
  | Equal
  | Geq
  | Greater
  | Leq
  | Less
  | Mod
  | Mult
  | NEqual
  | Or
  | Sub

type uop =
    Neg
  | Not

type typ = 
    Bool
  | EdgeTyp of string
  | Float
  | Int
  | ListTyp of typ
  | NodeTyp of string
  | PackTyp of string
  | Packet
  | Str
  | TupleTyp of string
  | Void

type bind = typ * string

type expr =
    Assign of expr * expr
  | Binop of expr * op * expr
  | BoolLit of bool
  | Call of string * expr list
  | FloatLit of float
  | Id of string
  | IntLit of int
  | Lst of typ * expr list
  | ListAdd of expr * expr
  | ListRmv of expr * expr
  | Member of expr * string
  | Message
  | Node of string * expr list
  | Graph of typ * (expr * expr * expr list) list
  | Noexpr
  | Null of typ
  | Pack of expr list
  | Print of string * expr list
  | Run of expr * expr list
  | Self
  | StrLit of string
  | Subscript of expr * expr
  | Tuple of string * expr list
  | Unop of uop * expr

type pass_typ =
    Chld
  | ChldPrnt
  | Prnt
  | PrntChld
  | Target of expr

type stmt =
    Block of stmt list
  | Expr of expr
  | For of expr * expr * expr * stmt
  | ForEach of expr * expr * stmt
  | If of expr * stmt * stmt 
  | Pass of expr * pass_typ
  | Return of expr 
  | While of expr * stmt

type func_decl = {
    f_typ : typ;
    f_name : string;
    f_formals : bind list;
    f_locals : bind list;
    f_body : stmt list;
}

type node_decl = {
    n_typ  : string;
    n_data : bind list;
    n_edge : bind list;
    n_pack : bind list;
    n_do : func_decl;
    n_catch : func_decl;
}

type tuple_decl = {
    t_typ : string;
    t_attributes : bind list;
}

type program = node_decl list * tuple_decl list * bind list * func_decl list

(* Pretty-printing functions *)
let rec string_of_typ = function
    Int            -> "int"
  | Bool           -> "bool"
  | Void           -> "void"
  | Str            -> "string"
  | Float          -> "float"
  | NodeTyp(name)  -> name
  | Packet         -> "Packet"
  | EdgeTyp(name)  -> "edge//"^name
  | PackTyp(name)  -> "pack//"^name
  | ListTyp(typ)   -> "list " ^ string_of_typ typ
  | TupleTyp(name) -> name

module Typ = 
  struct
    type t = typ
    let compare t1 t2 = 
      String.compare (string_of_typ t1) (string_of_typ t2)
    let equal t1 t2 = compare t1 t2 = 0
  end

let string_of_pass_typ = function
    Chld     -> "Child"
  | ChldPrnt -> "ChildParent"
  | Prnt     -> "Parent"
  | PrntChld -> "ParentChild"
  | Target(_)-> "Target"
 
let string_of_op = function
    Add       -> "+"
  | Sub       -> "-"
  | Mult      -> "*"
  | Div       -> "/"
  | Equal     -> "=="
  | NEqual    -> "!="
  | Less      -> "<"
  | Leq       -> "<="
  | Greater   -> ">"
  | Geq       -> ">="
  | And       -> "&&"
  | Or        -> "||"
  | Mod       -> "%"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let rec string_of_expr = function
  | Assign(e1, e2) -> string_of_expr e1 ^ " = " ^ string_of_expr e2
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | FloatLit(f) -> string_of_float f
  | Graph(_, _) -> "graph" 
  | Id(s) -> s
  | IntLit(i) -> string_of_int i
  | Lst(typ, el) -> "new " ^ string_of_typ typ ^ " [|" ^
      String.concat ", " (List.map string_of_expr el) ^ "|]"
  | ListAdd(e1, e2) -> string_of_expr e1 ^ " += " ^ string_of_expr e2
  | ListRmv(e1, e2) -> string_of_expr e1 ^ " -= " ^ string_of_expr e2
  | Member(e, id) -> string_of_expr e ^ "." ^ id
  | Message -> "message"
  | Node(typ, el) -> typ ^ "(" ^
      String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | Null(typ) -> "null(" ^ string_of_typ typ ^ ")"
  | Pack(el) -> String.concat "\n" (List.map string_of_expr el)
  | Print(s, al) -> "print(\"" ^ String.escaped s ^ "\"" ^ if List.length al > 0
      then ", " ^ String.concat ", " (List.map string_of_expr al) ^ ")" else ")"
  | Run(e, bl) -> "run{" ^ string_of_expr e ^ "}(" ^
      String.concat ", " (List.map string_of_expr bl) ^ ")"
  | Self -> "self"
  | StrLit(s) -> s
  | Subscript(e1, e2) -> string_of_expr e1 ^ "[" ^ string_of_expr e2 ^ "]"
  | Tuple(typ, el) -> typ ^ "(" ^
      String.concat ", " (List.map string_of_expr el) ^ ")"
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e

let rec string_of_stmt = function
    Expr(e)
      -> string_of_expr e ^ ";"
  | Block(stmts)
      -> "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "\n}"
  | Return(expr)
      -> "return " ^ string_of_expr expr ^ ";"
  | If(e, s1, s2)
      -> "if (\n" ^ string_of_expr e ^ "\n)\n" ^ string_of_stmt s1 ^
         "\nelse\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s)
      -> "for(" ^ string_of_expr e1  ^ "; " ^ string_of_expr e2 ^ "; " ^
         string_of_expr e3  ^ ") " ^ string_of_stmt s
  | ForEach(e1, e2, s)
      -> "for(" ^ string_of_expr e1 ^ " in " ^ string_of_expr e2 ^ ") " ^
         string_of_stmt s
  | While(e, s)
      -> "while(" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | Pass (e, ptyp)
      -> "Pass(" ^ string_of_expr e ^ ")\n" ^ string_of_pass_typ ptyp

let string_of_bind (typ, id) =
  string_of_typ typ ^ " " ^ id

let string_of_formals formals =
  String.concat ", " (List.map string_of_bind formals)

let string_of_bind_list blist indent =
  if List.length blist > 0
    then indent ^ String.concat (";\n" ^ indent) (List.map string_of_bind blist)
    ^ ";\n"
  else ""

let string_of_body body indent =
  if List.length body > 0
    then indent ^ String.concat ("\n" ^ indent) (List.map string_of_stmt body)
  else ""

let string_of_func_decl func_decl =
  string_of_typ func_decl.f_typ ^ " " ^ func_decl.f_name ^ "(" ^
  string_of_formals func_decl.f_formals ^ ") {\n" ^
  string_of_bind_list func_decl.f_locals "  " ^
  string_of_body func_decl.f_body "  " ^ "\n}"

let string_of_n_do n_do = 
  "do(" ^
  string_of_formals n_do.f_formals ^ ") {\n" ^
  string_of_bind_list n_do.f_locals "    " ^
  string_of_body n_do.f_body "    " ^ "\n  }"

let string_of_n_catch n_catch =
  string_of_bind_list n_catch.f_locals "    " ^
  string_of_body n_catch.f_body "    "

let string_of_node_decl node_decl =
  node_decl.n_typ ^ " {\n  data {\n" ^
  string_of_bind_list node_decl.n_data "    " ^ "  }\n  edge {\n" ^
  string_of_bind_list node_decl.n_edge "    " ^ "  }\n  pack {\n" ^
  string_of_bind_list node_decl.n_pack "    " ^ "  }\n  " ^
  string_of_n_do node_decl.n_do ^ "\n  catch {\n" ^
  string_of_n_catch node_decl.n_catch ^ "\n  }\n}"

let string_of_tuple_decl tuple_decl =
  tuple_decl.t_typ ^ " {\n" ^
  string_of_bind_list tuple_decl.t_attributes "  " ^ "}"

let string_of_program (nodes, tuples, vars, funcs) =
  (if List.length vars > 0 then 
  String.concat ";\n" (List.map string_of_bind (List.rev vars)) ^ ";\n\n"
  else "") ^ (if List.length tuples > 0 then
  String.concat "\n\n" (List.map string_of_tuple_decl (List.rev tuples)) ^ "\n\n"
  else "") ^ (if List.length nodes > 0 then
  String.concat "\n\n" (List.map string_of_node_decl (List.rev nodes)) ^ "\n\n"
  else "") ^ (if List.length funcs > 0 then
  String.concat "\n\n" (List.map string_of_func_decl (List.rev funcs))
  else "") ^ "\n"
