(* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)
module L = Llvm
module A = Ast

module StringMap = Map.Make(String)
module TypMap = Map.Make(A.Typ)

type f_cat = Func|Do|Catch
type bind = {typ : A.typ; idx : int} 
type tuple_info = {ltyp : L.lltype; attrs : bind StringMap.t}

let translate (nodes, tuples, globals, functions) =
  let context = L.global_context () in
  let llctx = L.global_context () in
  let the_module = L.create_module context "MicroC" in
  let listm = L.MemoryBuffer.of_file "list.bc" in
  let llm = Llvm_bitreader.parse_bitcode llctx listm in
  let i32_t  = L.i32_type  context
  and i8_t   = L.i8_type   context
  and i1_t   = L.i1_type   context
  and flt_t  = L.float_type context
  and void_t = L.void_type context 
  and idlist_t = L.pointer_type (match L.type_by_name llm "struct.IdList" with
    None -> raise (Invalid_argument "Option.get idlist")
  | Some x -> x) 
	and node_t = L.pointer_type (match L.type_by_name llm "struct.ListNode" with
		None -> raise (Invalid_argument "Option.get")
	| Some x -> x)  in

 	let update_typedef typ typedef = 
		match typ with
		  A.ListTyp _ 		-> typedef
		| A.NodeTyp _ | A.TupleTyp _ | A.PackTyp _ | A.EdgeTyp _ -> 
			if not (TypMap.mem typ typedef)
				then 
				let name = A.string_of_typ typ in
				let ltyp = L.named_struct_type context name in
				let typ_info = {ltyp = ltyp; attrs = StringMap.empty} in
				TypMap.add typ typ_info typedef
			else typedef
    | _ -> typedef in

	let ltype_of_typ typedef = function
      A.Int 	-> i32_t
    | A.Bool 	-> i1_t
    | A.Float -> flt_t
    | A.Void 	-> void_t 
    | A.Str 	-> L.pointer_type i8_t 
    | A.ListTyp _  -> idlist_t
    | typ -> L.pointer_type (TypMap.find typ typedef).ltyp in

  let (typedefs, obj_funs, do_decls, catch_decls) = 
		let make_attr (lst,attrs,typedef,i) (t,n) =
      let new_typedef = 
				match t with
					A.Int | A.Float | A.Bool | A.Void -> typedef
				| _ -> update_typedef t typedef in
			let new_attrs = StringMap.add n {typ=t; idx=i} attrs in 
			let new_l = (ltype_of_typ new_typedef t)::lst in
			(new_l, new_attrs, new_typedef, i+1) in

    let make_tuple_info typedef ltyp attrs = 
      let (new_l,new_attrs,_,_) = 
        List.fold_left make_attr ([],StringMap.empty,typedef,0) attrs in
      ignore(L.struct_set_body ltyp (Array.of_list (List.rev new_l)) false);
      {ltyp = ltyp; attrs = new_attrs} in

  	let tuple_typedefs = 
	  	let add_tuple_info typedef tdecl = 
        let typ_name = A.TupleTyp tdecl.A.t_typ in
        let typedef = if TypMap.mem typ_name typedef 
          then typedef else update_typedef typ_name typedef in
        let ltyp = (TypMap.find typ_name typedef).ltyp in
	  		(* Adding attribute *)
        TypMap.add typ_name (make_tuple_info typedef ltyp tdecl.A.t_attributes) typedef
	  	in List.fold_left add_tuple_info TypMap.empty (List.rev tuples) in

	  let (node_typedefs, obj_funs, do_decls, catch_decls) = 
      let add_node_info (typedef, obj_funs, do_decls, catch_decls) ndecl = 
        let n_typ_name = A.NodeTyp ndecl.A.n_typ in
        let p_typ_name = A.PackTyp ndecl.A.n_typ in
        let e_typ_name = A.EdgeTyp ndecl.A.n_typ in

        let (n_edge,n_pack,n_do,n_catch) = 
          (ndecl.A.n_edge,ndecl.A.n_pack,ndecl.A.n_do,ndecl.A.n_catch) in
	  		(* Has this been forward declared? *)
        let typedef = if TypMap.mem n_typ_name typedef 
          then typedef else update_typedef n_typ_name typedef in
        let ltyp = (TypMap.find n_typ_name typedef).ltyp in

        let edge_ltyp = L.named_struct_type context ("edge//"^ndecl.A.n_typ) in
        let edge_info = make_tuple_info typedef edge_ltyp n_edge in

        let pack_ltyp = L.named_struct_type context ("pack//"^ndecl.A.n_typ) in
        let pack_info = make_tuple_info typedef pack_ltyp n_pack in

        let typedef = TypMap.add e_typ_name edge_info typedef in
        let typedef = TypMap.add p_typ_name pack_info typedef in

        (* Declaring the DO & CATCH functions *)
        let obj_fun_decls fdecl = 
          let name = fdecl.A.f_name
          and formal_types = 
            Array.of_list (List.map (fun (t,_) -> ltype_of_typ typedef t) fdecl.A.f_formals) in
          let ftyp = L.function_type (ltype_of_typ typedef fdecl.A.f_typ) formal_types in
          let fdef = L.define_function name ftyp the_module in 
        (ftyp, fdef) in

	  		let (do_typ,do_fun) = obj_fun_decls n_do in
	  		let (catch_typ,catch_fun) = obj_fun_decls n_catch in 

	  		(* Adding attributes *)
        let (new_l,new_attrs,new_typedef,_) = 
          List.fold_left make_attr 
          ([L.pointer_type catch_typ; 
            L.pointer_type do_typ; 
            idlist_t; 
            idlist_t],StringMap.empty,typedef,4) ndecl.A.n_data in

				ignore(L.struct_set_body ltyp (Array.of_list (List.rev new_l)) false);

				(TypMap.add n_typ_name {ltyp = ltyp; attrs = new_attrs} new_typedef,
          TypMap.add n_typ_name ((n_do,do_fun), (n_catch,catch_fun)) obj_funs,
          n_do::do_decls,
          n_catch::catch_decls)
      in List.fold_left add_node_info 
        (tuple_typedefs, TypMap.empty, [], []) (List.rev nodes)
	  in
    (node_typedefs, obj_funs, do_decls, catch_decls)  in
      
  (* Declare each global variable; remember its value in a map *)
  let global_vars =
    let global_var m (t, n) =
      let init = L.const_null (ltype_of_typ typedefs t)
      in StringMap.add n ((L.define_global n init the_module),t) m in
    List.fold_left global_var StringMap.empty globals in

  (* Declare printf(), which the print built-in function will call *)
  let printf_t    = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func = L.declare_function "printf" printf_t the_module in
  
  let initIdList_t  = L.function_type idlist_t [| |] in
  let initIdList_f  = L.declare_function "initIdList" initIdList_t the_module in
  let appendId_t    = L.function_type idlist_t [| idlist_t; L.pointer_type i8_t; L.pointer_type i8_t |] in
  let appendId_f    = L.declare_function "appendId" appendId_t the_module in
  let indexIdList_t = L.function_type node_t [| idlist_t; i32_t |] in
  let indexIdList_f = L.declare_function "indexIdList" indexIdList_t the_module in
  let removeIdList_t = L.function_type idlist_t [| idlist_t; L.pointer_type i8_t |] in
  let removeIdList_f = L.declare_function "removeIdList" removeIdList_t the_module in
  let findNodeId_t = L.function_type node_t [| idlist_t; L.pointer_type i8_t |] in
  let findNodeId_f = L.declare_function "findNodeId" findNodeId_t the_module in
  let isEmptyIdList_t = L.function_type i8_t [| idlist_t |] in 
  let isEmptyIdList_f = L.declare_function "isEmptyList" isEmptyIdList_t the_module in


  (* Define each function (arguments and return type) so we can call it *)
  let function_decls =
    let function_decl m fdecl =
      let name = fdecl.A.f_name
      and formal_types =
				Array.of_list (List.map (fun (t,_) -> ltype_of_typ typedefs t) fdecl.A.f_formals)	in 
			let ftype = L.function_type (ltype_of_typ typedefs fdecl.A.f_typ) formal_types
	    in StringMap.add name (L.define_function name ftype the_module, fdecl) m
	  in List.fold_left function_decl StringMap.empty functions in

  (* Fill in the body of the given function *)
  let build_function_body fcat fdecl =
    let (the_function, _) = match fcat with
      Func -> (fst (StringMap.find fdecl.A.f_name function_decls),A.Void)
    | Do -> let ((_,do_fun),_) = TypMap.find (fst (List.hd fdecl.A.f_formals)) obj_funs 
            in (do_fun, fst (List.hd fdecl.A.f_formals))
    | Catch -> let (_,(_,catch_fun)) = TypMap.find (fst (List.hd fdecl.A.f_formals)) obj_funs 
                in (catch_fun, fst (List.hd fdecl.A.f_formals))  
    in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let local_vars =
      let add_formal m (t, n) p = L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ typedefs t) n builder in
      	ignore (L.build_store p local builder);
      StringMap.add n (local,t) m in

      let add_local m (t, n) =
        let (ltyp,typ) = match t with
          A.Packet -> 
            let (_,node_typ) = StringMap.find "self" m in
            (match node_typ with
              A.NodeTyp n_str -> 
                let p_typ = A.PackTyp n_str in
                (ltype_of_typ typedefs p_typ,p_typ)
            | _ -> raise (Invalid_argument (A.string_of_typ node_typ)))
        | typ -> (ltype_of_typ typedefs typ,typ) in
      	let local_var = L.build_alloca ltyp n builder
      in StringMap.add n (local_var, typ) m in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.A.f_formals
          (Array.to_list (L.params the_function)) in
      List.fold_left add_local formals fdecl.A.f_locals in

    (* Return the value for a variable or formal argument *)
    let lookup n = try StringMap.find n local_vars
                   with Not_found -> StringMap.find n global_vars in

    let get_nstr typ = match typ with 
        A.NodeTyp nstr -> nstr | _ -> raise (Invalid_argument "Pack(nodetype)") in

    let build_tuple_ptr ltyp name act expr builder = 
      let actuals = List.map (fun e -> fst (expr builder e)) act in
      let tup_ptr = L.build_malloc ltyp name builder in
      let add_attr i actual = 
        let a_ptr = L.build_struct_gep tup_ptr i "act" builder in 
        ignore(L.build_store actual a_ptr builder); i+1 in
      ignore(List.fold_left add_attr 0 actuals);
      tup_ptr in

    let get_dest_binding expr builder e_d =
      match e_d with
        A.Id s -> lookup s 
      | A.Member (e, s) -> 
          let (e_val, e_typ) = expr builder e in
          let attr_binding = StringMap.find s (TypMap.find e_typ typedefs).attrs in 
          let idx = attr_binding.idx in
          (L.build_struct_gep e_val idx "ptr" builder, attr_binding.typ)
      | _ -> raise (Invalid_argument "dest_bining") in

    (* Construct code for an expression; return its value *)
    let rec expr builder = function
    	A.IntLit i 		-> (L.const_int i32_t i, A.Int)
    | A.FloatLit f 	-> (L.const_float flt_t f, A.Float)
    | A.BoolLit b 	-> (L.const_int i1_t (if b then 1 else 0), A.Bool)
    | A.StrLit s -> (* TODO *)
      let str_ptr = L.build_alloca (L.array_type i8_t (String.length s)) "str" builder in
      (str_ptr, A.Str)
    | (* TODO *) A.Noexpr -> (L.const_int i32_t 0, A.Void)
    | A.Id s 				-> 
    	let binding = lookup s
    	in (L.build_load (fst binding) s builder, snd binding)
    | A.Binop (e1, op, e2) ->
  	  let (e1',t1) = expr builder e1
  	  and (e2',_) = expr builder e2 in
      let (oper,typ) = (match op with
          A.Mod     -> (L.build_urem, A.Int)
        | A.And     -> (L.build_and, A.Bool)
        | A.Or      -> (L.build_or, A.Bool)
        | A.Add     -> 
          if t1 = A.Float
          then (L.build_fadd, A.Float)
          else (L.build_add, A.Int)
        | A.Sub     -> 
          if t1 = A.Float
          then (L.build_fsub, A.Float)
          else (L.build_sub, A.Int)
        | A.Mult    -> 
          if t1 = A.Float
          then (L.build_fmul, A.Float)         
          else (L.build_mul, A.Int)
        | A.Div     -> 
          if t1 = A.Float
          then (L.build_fdiv, A.Float)
          else (L.build_sdiv, A.Int)
        | A.Equal   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.Oeq, A.Bool)
          else (L.build_icmp L.Icmp.Eq, A.Bool)
        | A.NEqual   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.One, A.Bool)
          else (L.build_icmp L.Icmp.Ne, A.Bool)
        | A.Less   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.Olt, A.Bool)
          else (L.build_icmp L.Icmp.Slt, A.Bool)
        | A.Leq   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.Ole, A.Bool)
          else (L.build_icmp L.Icmp.Sle, A.Bool)
        | A.Greater   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.Ogt, A.Bool)
          else (L.build_icmp L.Icmp.Sgt, A.Bool)
        | A.Geq   -> 
          if t1 = A.Float
          then (L.build_fcmp L.Fcmp.Oge, A.Bool)
          else (L.build_icmp L.Icmp.Sge, A.Bool)) in
      (* ignore (print_endline (A.string_of_typ typ)); *)
  	  (oper e1' e2' "tmp" builder, typ)

    | A.ListAdd(l, d) ->
      let (l_val, l_typ) = expr builder l in
      let (d_val, d_typ) = expr builder d in
      let d_ptr = match d_typ with 
        A.NodeTyp _ | A.TupleTyp _ | A.PackTyp _ | A.EdgeTyp _-> d_val
      | A.ListTyp _ -> d_val
      | _ ->
        let d_ltyp = ltype_of_typ typedefs d_typ in
        let d_ptr = L.build_malloc d_ltyp "tmp" builder in
        ignore (L.build_store d_val d_ptr builder);
        d_ptr in

      let void_d_ptr = L.build_bitcast d_ptr (L.pointer_type i8_t) "ptr" builder in
      ignore(L.build_call appendId_f [| l_val; void_d_ptr; void_d_ptr |] "tmp" builder);
      (l_val, l_typ)

    | A.ListRmv(l, d) ->
      let (l_val, l_typ) = expr builder l in
      let (d_val, _) = expr builder d in
      let void_d_ptr = L.build_bitcast d_val (L.pointer_type i8_t) "ptr" builder in
      ignore (L.build_call removeIdList_f [| l_val; void_d_ptr |] "" builder);
      (l_val, l_typ)

    | A.Unop(op, e) ->
      let (e_val,e_typ) = expr builder e in
      (match e_typ with
        A.NodeTyp _ ->
          let chld_lst_ptr = L.build_struct_gep e_val 0 "chld_ptr" builder in
          let chld_lst = L.build_load chld_lst_ptr "chld" builder in
          let is_empty = L.build_call isEmptyIdList_f [| chld_lst |] "is_empty" builder  in
          (L.build_icmp L.Icmp.Ne is_empty (L.const_null i8_t) "is_empty_bool" builder, A.Bool)
      | _ ->
        ((match op with
                    A.Neg     -> L.build_neg
                  | A.Not     -> L.build_not) e_val "tmp" builder, A.Bool))
    | A.Member (e, s) -> 
      let (e_val,e_typ) = expr builder e in
      (* ignore(print_endline (L.string_of_lltype (L.type_of e_val))); *)
      let attr_binding = StringMap.find s (TypMap.find e_typ typedefs).attrs in 
      let idx = attr_binding.idx in
      let ptr = L.build_struct_gep e_val idx "ptr" builder in
      (L.build_load ptr "tmp" builder, attr_binding.typ)

  (* FIX *)
    | A.Assign (e_d, e_s) -> 
      let (lval,_) = expr builder e_s in
    	let (dst_val,dst_typ) = get_dest_binding expr builder e_d in
      ignore (L.build_store lval dst_val builder); (lval, dst_typ)
  			
  (* FIX *)
    | A.Print (fmt, act) ->
      let build_print_arg e = 
        let (e_val, e_typ) = expr builder e in
          match e_typ with 
            A.Float -> L.build_fpext e_val (L.double_type context) "dbl_cast" builder
          | _ -> e_val in
      let expr_lst = List.map build_print_arg act in
      let format_str = L.build_global_stringptr fmt "fmt" builder in
      (L.build_call printf_func (Array.of_list (format_str::expr_lst)) "printf" builder, A.Void)

    | A.Call (f, act) ->
      let (fdef, fdecl) = StringMap.find f function_decls in
      let actuals = List.rev (List.map (fun e -> fst (expr builder e)) (List.rev act)) in
      let result = (match fdecl.A.f_typ with A.Void -> ""
                                          | _ -> f ^ "_result") in
      (L.build_call fdef (Array.of_list actuals) result builder, fdecl.A.f_typ)

    | A.Run (e, act) ->
      let (n_val, n_typ) = expr builder e in
      let ((do_typ,do_fun),_) = TypMap.find n_typ obj_funs in
      let actuals = n_val::(List.rev (List.map (fun e -> fst (expr builder e)) (List.rev act))) in
      let result = (match do_typ.A.f_typ with A.Void -> ""
                                      | _ -> (A.string_of_typ n_typ)^"do_result" ) in
      (L.build_call do_fun (Array.of_list actuals) result builder, do_typ.A.f_typ)
 
    | A.Self -> expr builder (A.Id "self")

    | A.Message -> expr builder (A.Id "msg")

    | A.Null typ -> 
      let ltyp = ltype_of_typ typedefs typ in
      (L.const_null ltyp,typ)

    | A.Tuple (typ, act) ->
    	let ltyp = (TypMap.find (A.TupleTyp typ) typedefs).ltyp in
      (build_tuple_ptr ltyp "tuple" act expr builder, A.NodeTyp typ)

    | A.Node (typ, act) ->
      let ltyp = (TypMap.find (A.NodeTyp typ) typedefs).ltyp in
      let ((_,do_fun), (_,catch_fun)) = TypMap.find (A.NodeTyp typ) obj_funs in
      let actuals = List.append [L.build_call initIdList_f [||] "init" builder;
                                L.build_call initIdList_f [||] "init" builder;
                                do_fun; 
                                catch_fun;] 
                    (List.map (fun e -> fst (expr builder e)) act) in
      let node_ptr = L.build_malloc ltyp "tmp" builder in
      let add_attr i actual = 
        let a_ptr = L.build_struct_gep node_ptr i "act" builder in
        ignore(L.build_store actual a_ptr builder); i+1 in
      ignore(List.fold_left add_attr 0 actuals);
      (node_ptr, A.NodeTyp typ)

    | A.Lst (typ, act) -> (* (L.const_null i32_t, A.Void) *)
    	let d_ltyp = ltype_of_typ typedefs typ in
    	let lst_ptr = L.build_call initIdList_f [||] "init" builder in
    	let add_act dat = 
        let d_ptr = match typ with 
          A.NodeTyp _ | A.TupleTyp _ | A.PackTyp _ | A.EdgeTyp _-> 
            fst (expr builder dat)
        | A.ListTyp _ -> fst (expr builder dat)
        | _ ->
    		  let d_val = fst (expr builder dat) in
          let d_ptr = L.build_malloc d_ltyp "tmp" builder in
          ignore (L.build_store d_val d_ptr builder);
    		  d_ptr in

    		let void_d_ptr = L.build_bitcast d_ptr (L.pointer_type i8_t) "ptr" builder in
    		ignore (L.build_call appendId_f [| lst_ptr; void_d_ptr; void_d_ptr |] "tmp" builder) 
    	in ignore (List.map add_act act);
    	(lst_ptr, A.ListTyp typ)

    | A.Pack (act) ->
      let (_,n_typ) = lookup "self" in
      let n_str = get_nstr n_typ in
      let pack_info = TypMap.find (A.PackTyp n_str) typedefs in
      let pack_ltyp = pack_info.ltyp in
      (build_tuple_ptr pack_ltyp "msg" act expr builder, A.PackTyp(n_str))

    | A.Graph (n_typ, edge_lst) ->
      let n_str = get_nstr n_typ in
      let edge_info = TypMap.find (A.EdgeTyp n_str) typedefs in
      let edge_ltyp = edge_info.ltyp in
      let build_edges edge_decl = 
        let (src,dst,act_l) = edge_decl in
        let (src_val,_) = expr builder src in
        let (dst_val,_) = expr builder dst in 

        let edge = build_tuple_ptr edge_ltyp "edge" act_l expr builder in
        let void_edge = L.build_bitcast edge (L.pointer_type i8_t) "edge_ptr" builder in

        let chld_lst_ptr = L.build_struct_gep src_val 0 "chld_lst_ptr" builder in
        let chld_lst = L.build_load chld_lst_ptr "chld_lst" builder in
        let void_dst_val = L.build_bitcast dst_val (L.pointer_type i8_t) "dst_ptr" builder in
        ignore (L.build_call appendId_f [| chld_lst; void_edge; void_dst_val |] "append" builder);

        let prnt_lst_ptr = L.build_struct_gep dst_val 1 "prnt_lst_ptr" builder in
        let prnt_lst = L.build_load prnt_lst_ptr "prnt_lst" builder in
        let void_src_val = L.build_bitcast src_val (L.pointer_type i8_t) "src_ptr" builder in
        ignore (L.build_call appendId_f [| prnt_lst; void_edge; void_src_val |] "append" builder);

      in ignore(List.iter build_edges edge_lst); 
      (L.const_null i8_t, A.Void)

		| A.Subscript(e, sub) -> 
			let (sub_val,_) = expr builder sub in
      let (e_val,e_typ) = expr builder e in
      match e_typ with
        A.ListTyp typ ->
          let l_node_ptr = L.build_call indexIdList_f [| e_val; sub_val |] "node" builder in
          let l_node_data_ptr = L.build_struct_gep l_node_ptr 0 "l_node_data_ptr" builder in
          (match typ with 
            A.NodeTyp _ | A.TupleTyp _ | A.PackTyp _ | A.EdgeTyp _-> 
              let l_dtyp = ltype_of_typ typedefs typ in
              let void_d_ptr = L.build_load l_node_data_ptr "void_d_ptr" builder in
              (L.build_bitcast void_d_ptr l_dtyp "data" builder, typ)
          | A.ListTyp _ ->
              let l_dtyp = ltype_of_typ typedefs typ in
              let void_d_ptr = L.build_load l_node_data_ptr "void_d_ptr" builder in
              (L.build_bitcast void_d_ptr l_dtyp "data" builder, typ)
          | _ -> 
            let l_dtyp = ltype_of_typ typedefs typ in
            let void_d_ptr = L.build_load l_node_data_ptr "void_d_ptr" builder in
            let d_ptr = L.build_bitcast void_d_ptr (L.pointer_type l_dtyp) "d_ptr" builder in
            (L.build_load d_ptr "d_ptr" builder, typ))

      | A.NodeTyp n_str -> 
          let e_ltyp = ltype_of_typ typedefs (A.EdgeTyp n_str) in
          (* let edge_info = TypMap.find (A.EdgeTyp n_str) typedefs in *)
          let chld_lst_ptr = L.build_struct_gep e_val 0 "chdl_lst_ptr" builder in
          let chld_lst = L.build_load chld_lst_ptr "chld_lst" builder in

          let id_val = L.build_bitcast sub_val (L.pointer_type i8_t) "void_id_val" builder in
          let lst_node = L.build_call findNodeId_f [| chld_lst; id_val |] "lst_node" builder in
          let void_d_ptr = L.build_struct_gep lst_node 0 "lst_node" builder in
          (* let void_d_ptr = L.build_load data_ptr "void_d_ptr" builder in *)
          let data_ptr = L.build_bitcast void_d_ptr (L.pointer_type e_ltyp) "d_ptr" builder in
          (L.build_load data_ptr "data" builder, A.EdgeTyp(n_str))
      | typ -> raise (Invalid_argument (A.string_of_typ typ))
	  in

    let add_terminal builder f =
      match L.block_terminator (L.insertion_block builder) with
				Some _ -> ()
      | None -> ignore (f builder) in
	
    (* Build the code for the given statement; return the builder for
       the statement's successor *)
    let rec stmt builder = function
			A.Block sl -> List.fold_left stmt builder sl
    | A.Expr e -> ignore (expr builder e); builder
    | A.Return e -> ignore (match fdecl.A.f_typ with
  			A.Void -> L.build_ret_void builder
			| _ -> L.build_ret (fst (expr builder e)) builder); builder

    | A.If (predicate, then_stmt, else_stmt) ->
      let bool_val = fst (expr builder predicate) in
 			let merge_bb = L.append_block context "merge" the_function in

			let then_bb = L.append_block context "then" the_function in
			add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
			 (L.build_br merge_bb);

			let else_bb = L.append_block context "else" the_function in
			add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
			 (L.build_br merge_bb);

			ignore (L.build_cond_br bool_val then_bb else_bb builder);
			L.builder_at_end context merge_bb

    | A.While (predicate, body) ->
		  let pred_bb = L.append_block context "while" the_function in
		  ignore (L.build_br pred_bb builder);

		  let body_bb = L.append_block context "while_body" the_function in
		  add_terminal (stmt (L.builder_at_end context body_bb) body)
		    (L.build_br pred_bb);

		  let pred_builder = L.builder_at_end context pred_bb in
		  let bool_val = fst (expr pred_builder predicate) in

		  let merge_bb = L.append_block context "merge" the_function in
		  ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
		  L.builder_at_end context merge_bb

    | A.For (e1, e2, e3, body) -> stmt builder
	    ( A.Block [A.Expr e1 ; A.While (e2, A.Block [body ; A.Expr e3]) ] )

    | A.ForEach (s, l, body) ->
      let lst_node = L.build_alloca node_t "lst_node" builder in
      let (lst,typ) = expr builder l in
      let head_ptr_ptr = L.build_struct_gep lst 0 "head_ptr" builder in
      let head_ptr = L.build_load head_ptr_ptr "head" builder in
      ignore (L.build_store head_ptr lst_node builder);

      let pred_bb = L.append_block context "while" the_function in
      ignore (L.build_br pred_bb builder);

      let body_bb = L.append_block context "while_body" the_function in
      let body_builder = L.builder_at_end context body_bb in
      let (l_dtyp,dtyp) = match typ with
        A.ListTyp dtyp -> (ltype_of_typ typedefs dtyp, dtyp)
      | _ -> (void_t, A.Void) in

      let curr = L.build_load lst_node "curr" body_builder in
      let void_data_ptr_ptr = L.build_struct_gep curr 0 "void_data_ptr_ptr" body_builder in
      let void_data_ptr = L.build_load void_data_ptr_ptr "void_data_ptr" body_builder in
      let data_val = match dtyp with
        A.NodeTyp _ | A.TupleTyp _ | A.EdgeTyp _ 
          -> L.build_bitcast void_data_ptr l_dtyp "data" body_builder
      | A.ListTyp _
          -> L.build_bitcast void_data_ptr l_dtyp "data" body_builder
      | _ -> 
          let data_ptr = 
          L.build_bitcast void_data_ptr (L.pointer_type l_dtyp) "data_ptr" body_builder in
          L.build_load data_ptr "data" body_builder in
      let (dst_ptr,_) = get_dest_binding expr body_builder s in
      ignore(L.build_store data_val dst_ptr body_builder);

      let body_builder = stmt body_builder body in

      let next_ptr = L.build_struct_gep curr 2 "next_ptr" body_builder in
      let next = L.build_load next_ptr "next" body_builder in
      ignore(L.build_store next lst_node body_builder);
      add_terminal body_builder (L.build_br pred_bb);

      let pred_builder = L.builder_at_end context pred_bb in
      let bool_val = 
        let curr = L.build_load lst_node "curr" pred_builder in
        L.build_icmp L.Icmp.Ne curr (L.const_null node_t) "tmp" pred_builder in

      let merge_bb = L.append_block context "merge" the_function in
      ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
      L.builder_at_end context merge_bb

    | A.Pass (msg, p_typ) ->
      let (msg_val,_) = expr builder msg in

      let (self_ptr,n_typ) = (lookup "self") in
      let self = L.build_load self_ptr "self" builder in
      let (_,(_,catch_fun)) = TypMap.find n_typ obj_funs in
      let lst_node = L.build_alloca node_t "curr_node" builder in

      let chld_lst_ptr = L.build_struct_gep self 0 "lst_ptr" builder in
      let prnt_lst_ptr = L.build_struct_gep self 1 "lst_ptr" builder in

      let lst_iter_catch builder lst_ptr = 
        let lst = L.build_load lst_ptr "lst" builder in
        let head_ptr = L.build_struct_gep lst 0 "head_ptr" builder in
        let head = L.build_load head_ptr "head" builder in
        ignore(L.build_store head lst_node builder);
        let pred_bb = L.append_block context "while" the_function in
        ignore (L.build_br pred_bb builder);

        let body_bb = L.append_block context "while_body" the_function in
        let body_builder = L.builder_at_end context body_bb in

        let curr = L.build_load lst_node "curr" body_builder in
        let void_dst_ptr = L.build_struct_gep curr 1 "void_dst_ptr" body_builder in
        let void_dst = L.build_load void_dst_ptr "void_dst" body_builder in
        let dst = L.build_bitcast void_dst (ltype_of_typ typedefs n_typ) "dst" body_builder in
        ignore (L.build_call catch_fun [| dst; msg_val |] "" body_builder);

        let next_ptr = L.build_struct_gep curr 2 "next_ptr" body_builder in
        let next = L.build_load next_ptr "next" body_builder in
        ignore(L.build_store next lst_node body_builder);
        add_terminal body_builder (L.build_br pred_bb);

        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = 
          let curr = L.build_load lst_node "curr" pred_builder in
          L.build_icmp L.Icmp.Ne curr (L.const_null node_t) "tmp" pred_builder in

        let merge_bb = L.append_block context "merge" the_function in
        ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb
      in 

      match p_typ with
        A.Prnt      -> List.fold_left lst_iter_catch builder [prnt_lst_ptr]
      | A.Chld      -> List.fold_left lst_iter_catch builder [chld_lst_ptr]
      | A.PrntChld  -> List.fold_left lst_iter_catch builder [prnt_lst_ptr; chld_lst_ptr]
      | A.ChldPrnt  -> List.fold_left lst_iter_catch builder [chld_lst_ptr; prnt_lst_ptr] 
      | A.Target dst  -> 
          let (dst_val,_) = expr builder dst in
          ignore(L.build_call catch_fun [| dst_val; msg_val |] "" builder); builder

    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (A.Block fdecl.A.f_body) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.A.f_typ with
        A.Void -> L.build_ret_void
      | t -> L.build_ret (L.const_int (ltype_of_typ typedefs t) 0))
  in
  List.iter (build_function_body Do) do_decls;
  List.iter (build_function_body Catch) catch_decls;
  List.iter (build_function_body Func) functions;
the_module
