/* Ocamlyacc parser for GOBLAN */

%{
open Ast;;
let parse_error s = (* Called by the parser function on error *)
  print_endline s;
  flush stdout;;
let get1 (a,_,_,_) = a;;
let get2 (_,a,_,_) = a;;
let get3 (_,_,a,_) = a;;
let get4 (_,_,_,a) = a;;

let add_self n_typ fdecl =
  { f_typ     = fdecl.f_typ;
    f_name    = n_typ ^ "." ^ fdecl.f_name;
    f_formals = (NodeTyp(n_typ), "self")::fdecl.f_formals;
    f_locals  = fdecl.f_locals;
    f_body    = fdecl.f_body }

let add_self_pack n_typ fdecl = 
  { f_typ     = fdecl.f_typ;
    f_name    = n_typ ^ "." ^ fdecl.f_name;
    f_formals = (NodeTyp(n_typ), "self")::(PackTyp(n_typ), "msg")::fdecl.f_formals;
    f_locals  = fdecl.f_locals;
    f_body    = fdecl.f_body }
%}

%token LPAREN RPAREN LBRACE RBRACE LLIST RLIST LBRACKET RBRACKET
%token SEMI COMMA PERIOD PLUS MINUS TIMES DIVIDE MODULO
%token ASSIGN EQ NEQ LT LEQ GT GEQ AND OR NOT ARROW LSTADD LSTRMV
%token IF ELSE FOR IN WHILE RETURN PRINT
%token BOOL INT FLOAT STRING LIST NEW VOID EDGE PACK GRAPH
%token TRUE FALSE DATA DO CATCH SELF PARENT CHILD PRNTCHLD CHLDPRNT MESSAGE
%token PASS RUN NULL LSTADD LSTRMV

%token <string> ID
%token <float>  FLT_LIT
%token <int>    INT_LIT
%token <string> STR_LIT
%token <string> NODE_TYP TUPLE_TYP
%token <string> STR
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left LSTADD LSTRMV
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MODULO
%nonassoc LBRACKET RBRACKET
%left NOT NEG
%left PERIOD

%start program
%type <Ast.program> program

%%

program:
  decls EOF                            { $1 }

decls:
   /* nothing */                       { [], [], [], [] }
 | decls ndecl                         { ($2 :: get1 $1), get2 $1, get3 $1, get4 $1 }
 | decls tdecl                         { get1 $1, ($2 :: get2 $1), get3 $1, get4 $1 }
 | decls vdecl                         { get1 $1, get2 $1, ($2 :: get3 $1), get4 $1 }
 | decls fdecl                         { get1 $1, get2 $1, get3 $1, ($2 :: get4 $1) }

vdecl:
    typ ID SEMI { ($1, $2) }

vdecl_list:
    /* nothing */                      { [] }
  | vdecl_list vdecl                   { $2 :: $1 }

typ:
    INT                                { Int }
  | BOOL                               { Bool }
  | VOID                               { Void }
  | STRING                             { Str }
  | FLOAT                              { Float }
  | NODE_TYP                           { NodeTyp($1) }
  | PACK                               { Packet }
  | LIST typ                           { ListTyp($2) }
  | TUPLE_TYP                          { TupleTyp($1) }

ndecl:
   NODE_TYP LBRACE n_data n_edge n_pack n_do n_catch RBRACE
                                       { { n_typ = $1;
                                           n_data = $3;
                                           n_edge = $4;
                                           n_pack = $5;
                                           n_do = add_self $1 $6;
                                           n_catch = add_self_pack $1 $7 } }

n_data:
   DATA LBRACE vdecl_list RBRACE       { List.rev $3 }

n_edge:
   EDGE LBRACE vdecl_list RBRACE       { List.rev $3 }

n_pack:
   PACK LBRACE vdecl_list RBRACE       { List.rev $3 }

n_do:
   typ DO LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
                                       { { f_typ = $1;
                                           f_name = "do";
                                           f_formals = $4;
                                           f_locals = (List.rev $7);
                                           f_body = List.rev $8 } }

n_catch:
   CATCH LBRACE vdecl_list stmt_list RBRACE
                                       { { f_typ = Void;
                                           f_name = "catch";
                                           f_formals = [];
                                           f_locals = List.rev $3;
                                           f_body = List.rev $4 } }

tdecl:
    TUPLE_TYP LBRACE vdecl_list RBRACE { { t_typ = $1;
                                           t_attributes = List.rev $3 } }

fdecl:
   typ ID LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
                                      { { f_typ = $1;
                                          f_name = $2;
                                          f_formals = $4;
                                          f_locals = List.rev $7;
                                          f_body = List.rev $8 } }
formals_opt:
    /* nothing */                      { [] }
  | formal_list                        { List.rev $1 }

formal_list:
    typ ID                             { [($1,$2)] }
  | formal_list COMMA typ ID           { ($3,$4) :: $1 }

stmt_list:
    /* nothing */                      { [] }
  | stmt_list stmt                     { $2 :: $1 }

pass_typ:
    PARENT    { Prnt }
  | CHILD     { Chld }
  | PRNTCHLD  { PrntChld }     
  | CHLDPRNT  { ChldPrnt }     
  | expr      { Target($1) }

stmt:
    expr SEMI                          { Expr $1 }
  | LBRACE stmt_list RBRACE            { Block(List.rev $2) }
  | RETURN SEMI                        { Return Noexpr }
  | RETURN expr SEMI                   { Return $2 }
  | IF LPAREN expr RPAREN stmt %prec NOELSE
                                       { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt 
                                       { If($3, $5, $7) }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
                                       { For($3, $5, $7, $9) }
  | FOR LPAREN expr IN expr RPAREN stmt
                                       { ForEach($3, $5, $7)}
  | WHILE LPAREN expr RPAREN stmt
                                       { While($3, $5) }
  | PASS expr ARROW pass_typ SEMI      { Pass($2, $4) }
  
expr_opt:
    /* nothing */                      { Noexpr }
  | expr                               { $1 }

expr:
    expr ASSIGN  expr                  { Assign($1, $3) }
  | expr PLUS    expr                  { Binop($1, Add, $3) }
  | expr MINUS   expr                  { Binop($1, Sub, $3) }
  | expr TIMES   expr                  { Binop($1, Mult, $3) }
  | expr DIVIDE  expr                  { Binop($1, Div, $3) }
  | expr MODULO  expr                  { Binop($1, Mod, $3) }
  | expr EQ      expr                  { Binop($1, Equal, $3) }
  | expr NEQ     expr                  { Binop($1, NEqual, $3) }
  | expr LT      expr                  { Binop($1, Less, $3) }
  | expr LEQ     expr                  { Binop($1, Leq, $3) }
  | expr GT      expr                  { Binop($1, Greater, $3) }
  | expr GEQ     expr                  { Binop($1, Geq, $3) }
  | expr AND     expr                  { Binop($1, And, $3) }
  | expr OR      expr                  { Binop($1, Or, $3) }
  | expr LSTADD  expr                  { ListAdd($1, $3) }
  | expr LSTRMV  expr                  { ListRmv($1, $3) }
  | TRUE                               { BoolLit(true) }
  | FALSE                              { BoolLit(false) }
  | ID LPAREN actuals_opt RPAREN       { Call($1, $3) }
  | FLT_LIT                            { FloatLit($1) }
  | ID                                 { Id($1) }
  | INT_LIT                            { IntLit($1) }
  | NEW typ LLIST actuals_opt RLIST    { Lst($2, $4) }
  | expr PERIOD  ID                    { Member($1, $3) }
  | MESSAGE                            { Message }
  | NEW NODE_TYP LPAREN actuals_opt RPAREN
                                       { Node($2, $4) }
  | NEW GRAPH LPAREN NODE_TYP RPAREN LLIST edge_opt RLIST
                                       { Graph(NodeTyp($4), $7) }
  | NEW PACK LPAREN actuals_opt RPAREN { Pack($4) }
  | NULL LPAREN typ RPAREN             { Null($3) }
  | PRINT LPAREN STR_LIT print_actuals RPAREN 
                                       { Print($3, List.rev $4) }
  | RUN LBRACE expr RBRACE LPAREN actuals_opt RPAREN 
                                       { Run ($3,$6) }
  | SELF                               { Self }
  | STR_LIT                            { StrLit($1) }
  | expr LBRACKET expr RBRACKET        { Subscript($1, $3) }
  | NEW TUPLE_TYP LPAREN actuals_opt RPAREN
                                       { Tuple($2, $4) }
  | MINUS expr %prec NEG               { Unop(Neg, $2) }
  | NOT          expr                  { Unop(Not, $2) }
  | LPAREN expr RPAREN                 { $2 }  

edge_opt:
    /* nothing */                      { [] }
  | edge_list                          { List.rev $1 }

edge_list:
    edge_decl                          { [$1] }
  | edge_list COMMA edge_decl          { $3 :: $1 }

edge_decl:
  EDGE LBRACKET expr ARROW expr RBRACKET LPAREN actuals_opt RPAREN
                                       { ($3,$5,$8) }

print_actuals:
    /* nothing */                      { [] }
  | COMMA actuals_list                 { $2 }

actuals_opt:
    /* nothing */                      { [] }
  | actuals_list                       { List.rev $1 }

actuals_list:
    expr                               { [$1] }
  | actuals_list COMMA expr            { $3 :: $1 }
