(* GOBLAN Semantic Checker *)

open Ast
module StringMap = Map.Make(String)
module TypMap = Map.Make(Typ)

let check (nodes, tuples, globals, functions) =

(* Declaring typedef  *********************************************************)

  let add_to_attr attr_dict (t,n) =
    StringMap.add n t attr_dict in

  let typedefs = 
    let tuple_typedefs = 
      let add_tuple_attr typedef tdecl = 
        let typ_name = TupleTyp tdecl.t_typ in
        let attr_dict = 
          List.fold_left add_to_attr StringMap.empty tdecl.t_attributes in
        TypMap.add typ_name attr_dict typedef in
      List.fold_left add_tuple_attr TypMap.empty (List.rev tuples) in

    let node_typedefs = 
      let add_node_attr typedef ndecl = 
        let n_typ_name = NodeTyp ndecl.n_typ in
        let p_typ_name = PackTyp ndecl.n_typ in
        let e_typ_name = EdgeTyp ndecl.n_typ in
        let (n_data,n_edge, n_pack) = 
          (ndecl.n_data, ndecl.n_edge, ndecl.n_pack) in
        let n_attr = List.fold_left add_to_attr StringMap.empty n_data in
        let e_attr = List.fold_left add_to_attr StringMap.empty n_edge in
        let p_attr = List.fold_left add_to_attr StringMap.empty n_pack in
        
        let typedef = TypMap.add n_typ_name n_attr typedef in
        let typedef = TypMap.add e_typ_name e_attr typedef in
        TypMap.add p_typ_name p_attr typedef in
      List.fold_left add_node_attr tuple_typedefs (List.rev nodes) 
    in node_typedefs
  in
 
(******************************************************************************)

  (* Raise an exception if the given list has a duplicate *)
  let report_duplicate exceptf list =
    let rec helper = function
	    n1 :: n2 :: _ when n1 = n2 -> raise (Failure (exceptf n1))
        | _ :: t -> helper t
        | [] -> ()
    in helper (List.sort compare list)
  in

  (* Raise an exception if a given binding is to a void type *)
  let check_not_void exceptf = function
      (Void, n) -> raise (Failure (exceptf n))
    | _ -> ()
  in

  (* Raise an exception if the binding is to a PackTyp *)
  let check_not_packet exceptf = function
      (Packet, n) -> raise (Failure (exceptf n))
    | _ -> ()
  in

  (* Raise an exception of the given rvalue type cannot be assigned to
     the given lvalue type *)
  let check_assign lvaluet rvaluet err =
     if lvaluet = rvaluet then lvaluet else raise err
  in

  (* Raies an exception if the node or tuple type is not in typedefs *)
  let check_defined_nt typ instr =
    (match typ with
        NodeTyp(_) | TupleTyp(_) ->
          if not (TypMap.mem typ typedefs) then
            raise (Failure ("In " ^ instr ^ ", undefined " ^
            "reference to " ^ string_of_typ typ))
          else
            ()
      | _ -> ()
    )
  in

(* Checking Global Variables **************************************************)

  (* Checking globals for undefined references to tuples or nodes *)
  List.iter (fun (typ, _) -> check_defined_nt typ "globals"; ()) globals;

  (* Check global variables for void types *)
  List.iter (check_not_void (fun n -> "Illegal void global " ^ n)) globals;

  (* Check global variables for pack types *)
  List.iter (check_not_packet (fun n -> "Illegal pack global " ^ n)) globals;
   
  (* Check global variables for duplicate names *)
  report_duplicate (fun n -> "Duplicate global " ^ n) (List.map snd globals);

(* Checking Tuples ************************************************************)

  (* Check that there are no duplicate tuple types *)
  report_duplicate (fun n -> "Duplicate tuple " ^ n) (List.map (fun tuple ->
    tuple.t_typ) tuples);

  (* A function that is used to check each tuple (tuple) in the tuples_decl
     list *)
  let check_tuple tuple =

    (* Checks for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ tuple.t_typ; ())
    tuple.t_attributes;

    (* Checks for void types *)
    List.iter (check_not_void (fun n -> "Illegal void attribute " ^ n ^
      " in " ^ tuple.t_typ)) tuple.t_attributes;

    (* Checks for pack types *)
    List.iter (check_not_packet (fun n -> "Illegal pack attribute " ^ n ^
      " in " ^ tuple.t_typ)) tuple.t_attributes;
    
    (* Checks for duplicates *)
    report_duplicate (fun n -> "Duplicate tuple attribute " ^ n) (List.map (fun
      tuple_attr -> snd tuple_attr) tuple.t_attributes);

  in

  (* Call check_tuple on each tuple in the tuple list *)
  List.iter check_tuple tuples;

(* Checking Functions *********************************************************)

  (* Check that a function named main is defined *)
  if not (List.mem "main" (List.map (fun fd -> fd.f_name) functions))
  then raise (Failure ("Main function not found")) else ();

  (* Check that a function named print is not defined *)
  if List.mem "print" (List.map (fun fd -> fd.f_name) functions)
  then raise (Failure ("Function print may not be defined")) else ();

  (* Check that a function named node is not defined *)
  if List.mem "node" (List.map (fun fd -> fd.f_name) functions)
  then raise (Failure ("Function node may not be defined")) else ();

  (* Check that a function named tuple is not defined *)
  if List.mem "tuple" (List.map (fun fd -> fd.f_name) functions)
  then raise (Failure ("Function tuple may not be defined")) else ();

  (* Check that there are no duplicate function names *)
  report_duplicate (fun n -> "Duplicate function " ^ n)
    (List.map (fun fd -> fd.f_name) functions);

  (* Builds a string map (f_name --> func_decl) that contains all of the
     functions built in to our language *)
  let built_in_decls = StringMap.singleton "print"
    { f_typ = Void; f_name = "print"; f_formals = [(Int, "x")];
      f_locals = []; f_body = [] }
  in
  
  (* Builds a string map (f_name --> func_decl) that contains all of the
     functions in the program, including built in functions *)
  let function_decls = List.fold_left (fun m fd -> StringMap.add fd.f_name fd m)
    built_in_decls functions
  in

  (* A function that takes f_name, looks up the f_name in the 
     f_name --> func_decl string map, and returns the func_decl, or an exception
     if the f_name is not found in the string map *)
  let function_decl s = try StringMap.find s function_decls
    with Not_found -> raise (Failure ("Unrecognized function " ^ s))
  in

  (* A function that is used to check each function (func) in the functions
     func_decl list *)
  let check_function func =

    (* Checks formals for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ func.f_name; ())
    func.f_formals;

    (* Checks locals for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ func.f_name; ())
    func.f_locals;
 
    (* Checks formals for void types *)
    List.iter (check_not_void (fun n -> "In " ^ func.f_name ^ ", illegal void" ^
      " formal " ^ n)) func.f_formals;

    (* Checks locals for void types *)
    List.iter (check_not_void (fun n -> "In " ^ func.f_name ^ ", illegal void" ^
      " local " ^ n)) func.f_locals;

    (* Checks formals for type pack *)
    List.iter (check_not_packet (fun n -> "In " ^ func.f_name ^ ", illegal pack"
      ^ " formal " ^ n)) func.f_formals;

    (* Checks locals for type pack *)
    List.iter (check_not_packet (fun n -> "In " ^ func.f_name ^ ", illegal pack"
      ^ " local " ^ n)) func.f_locals;

    (* Checks for duplicates in formals *)
    report_duplicate (fun n -> "In " ^ func.f_name ^ ", duplicate formal " ^ n)
      (List.map snd func.f_formals);

    (* Checks for duplicates in locals *)
    report_duplicate (fun n -> "In " ^ func.f_name ^ ", duplicate local " ^ n)
      (List.map snd func.f_locals);

    (* Builds a string map (s_name --> s_typ) *)
    let symbols = List.fold_left (fun strmap (var_type, var_name) ->
      StringMap.add var_name var_type strmap) StringMap.empty (globals @
        func.f_formals @ func.f_locals)
    in

    (* A function that takes a s_name, looks up the s_name in the
       s_name --> s_typ string map, and returns the s_typ, or an exception
       if the s_name is not found in the string map *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("In " ^ func.f_name ^ ", undeclared " ^
      "identifier " ^ s))
    in

    (* A function that returns the type of the member *)
    let type_of_member e s =
      let e_typ = e in
      let attr_dict = try TypMap.find e_typ typedefs 
      with Not_found -> raise (Failure ("In " ^ func.f_name ^ ", " ^
      string_of_typ e_typ ^ " type not found")) in
      try StringMap.find s attr_dict 
      with Not_found -> raise (Failure ("In " ^ func.f_name ^ ", " ^
      s ^ " member not found"))
    in     

    (* A recurive function that validates returns the type of an expression, or
       throws an exception if the expression is invalid *)
    let rec expr = function
        Assign(e1, e2) as exp -> let lt = 
          expr e1 and rt = expr e2 in
            check_assign lt rt (Failure ("In " ^ func.f_name ^ ", illegal " ^
            "assignment " ^ string_of_typ lt ^ " = " ^ string_of_typ rt ^ " in "
            ^ string_of_expr exp))
      | Binop(e1, op, e2) as e -> let t1 = expr e1 and t2 = expr e2 in
        (match op with
            Add       when t1 = Float && t2 = Float -> Float
          | Add       when t1 = Int && t2 = Int -> Int
          | And       when t1 = Bool && t2 = Bool -> Bool
          | Div       when t1 = Float && t2 = Float -> Float
          | Div       when t1 = Int && t2 = Int -> Int
          | Equal     when t1 = t2 -> Bool
          | Geq       when t1 = Float && t2 = Float -> Bool
          | Geq       when t1 = Int && t2 = Int -> Bool
          | Greater   when t1 = Float && t2 = Float -> Bool
          | Greater   when t1 = Int && t2 = Int -> Bool
          | Leq       when t1 = Float && t2 = Float -> Bool
          | Leq       when t1 = Int && t2 = Int -> Bool
          | Less      when t1 = Float && t2 = Float -> Bool
          | Less      when t1 = Int && t2 = Int -> Bool
          | Mod       when t1 = Int && t2 = Int -> Int
          | Mult      when t1 = Float && t2 = Float -> Float
          | Mult      when t1 = Int && t2 = Int -> Int
          | NEqual    when t1 = t2 -> Bool
          | Or        when t1 = Bool && t2 = Bool -> Bool
          | Sub       when t1 = Float && t2 = Float -> Float
          | Sub       when t1 = Int && t2 = Int -> Int
          | _ -> raise (Failure ("In " ^ func.f_name ^ ", illegal binary " ^
            "operator " ^ string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
            string_of_typ t2 ^ " in " ^ string_of_expr e)))
      | BoolLit(_) -> Bool
      | Call(s, el) as call -> let func_decl = function_decl s in
        if List.length el <> List.length func_decl.f_formals then
          raise (Failure ("In " ^ func.f_name ^ ", expecting " ^ string_of_int 
          (List.length func_decl.f_formals) ^ " arguments in call to " ^
          string_of_expr call))
        else
          List.iter2 (fun (formal, _) exp -> let actual = expr exp in
            ignore (check_assign formal actual (Failure ("In " ^ func.f_name ^
            ", illegal actual argument " ^ string_of_typ actual ^ ", expected " ^
            string_of_typ formal ^ " in parameter " ^ string_of_expr exp ^
            " in call to " ^ string_of_expr call))))
            func_decl.f_formals el; func_decl.f_typ
      | FloatLit(_) -> Float
      | Graph(_, _) -> Void
      | Id(s) -> type_of_identifier s
      | IntLit(_) -> Int
      | Lst (t, _) -> ListTyp(t)
      | ListAdd(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in 
              if t <> ele then
                raise (Failure ("In " ^ func.f_name ^ ", illegal list addition, "
                ^ string_of_typ lst ^ " += " ^ string_of_typ ele ^ " in " ^
                string_of_expr e1 ^ " += " ^ string_of_expr e2))
              else
                ListTyp(t)
          | _ -> raise (Failure ("In " ^ func.f_name ^ ", expected left side " ^
            "of operator to be of type list, in " ^ string_of_expr e1 ^ " += " ^
            string_of_expr e2)))
      | ListRmv(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in (match t with
                NodeTyp(_) | TupleTyp(_) ->
                  if t <> ele then
                    raise (Failure ("In " ^ func.f_name ^ ", type mismatch in "
                    ^ "list removal, " ^ string_of_typ lst ^ " -= " ^
                    string_of_typ ele ^ " in " ^ string_of_expr e1 ^ " -= " ^
                    string_of_expr e2))
                  else
                    ListTyp(t)
              | _ -> raise (Failure ("In " ^ func.f_name ^ ", illegal list " ^
                "removal of non-mutable list type, " ^ string_of_typ lst ^ " -= "
                ^ string_of_typ ele ^ " in " ^ string_of_expr e1 ^ " -= " ^
                string_of_expr e2)))
          | _ -> raise (Failure ("In " ^ func.f_name ^ ", expected left side of"
            ^ " operator to be of type list, in " ^ string_of_expr e1 ^ " -= " ^
            string_of_expr e2))) 
      | Member(e, s) -> type_of_member (expr e) s
      | Message -> raise (Failure ("In " ^ func.f_name ^ ", illegal use of " ^
          "msg"))
      | Node(s, _) -> NodeTyp(s)
      | Noexpr -> Void
      | Null(t) -> t
      | Pack(_) -> raise (Failure ("In " ^ func.f_name ^ ", illegal use of " ^
          "pack"))
      | Print(_, _) -> Void
      | Run(n, _) -> let n_decl = try (List.find (fun node_decl ->
        node_decl.n_typ = string_of_typ (expr n)) nodes) with Not_found ->
        raise (Failure ("In " ^ func.f_name ^ ", did not find " ^
        string_of_expr n)) in n_decl.n_do.f_typ
      | Self -> raise (Failure ("In " ^ func.f_name ^ ", illegal use of self"))
      | StrLit(_) -> Str
      | Subscript(e1, e2) -> let ltyp = expr e1 in
        (match ltyp with
            ListTyp(t) -> let rtyp = expr e2 in
              (match rtyp with
                  Int -> t
                | _ -> raise (Failure ("In " ^ func.f_name ^ ", expected " ^
                  "integer index type for list subscript access, in " ^
                  string_of_expr e1 ^ "[ " ^
                  string_of_expr e2 ^ " ]")))
          | NodeTyp(t1) -> let rtyp = expr e2 in
              (match rtyp with
                  NodeTyp(t2) when t1 = t2 -> EdgeTyp(t2)
                | _ -> raise (Failure ("In " ^ func.f_name ^ ", expected node" ^
                  " index type for node subscript access, in " ^
                  string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ]")))
          | _ -> raise (Failure ("In " ^ func.f_name ^ ", expected left side " ^
            "of operator to be of type list or node, in " ^ string_of_expr e1 ^
            "[ " ^ string_of_expr e2 ^ " ] ")))
      | Tuple(s, _) -> TupleTyp(s)
      | Unop(uop, e) as ex -> let t = expr e in
        (match uop with
            Neg when t = Int -> Int
          | Not -> (match t with
          	| Bool -> Bool
    		| NodeTyp _ -> Bool 
    		| _ -> raise (Failure ("In " ^ func.f_name ^ ", illegal unary " ^
                  "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
                      string_of_expr ex)))
          | _ -> raise (Failure ("In " ^ func.f_name ^ ", illegal unary " ^
            "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
            string_of_expr ex)))
    in

    (* A function that checks if the expression is a boolean type, and throws
       an exception if it isn't *)
    let check_bool_expr e =
      if expr e <> Bool then
        raise (Failure ("In " ^ func.f_name ^ ", expected a boolean " ^
        "expression, in " ^ string_of_expr e))
      else ()
    in

    (* A recursive function that validates a statement, and throws an
       exception if the statment is invalid *)
    let rec stmt = function
        Block sl -> let rec check_block = function
            [Return _ as s] -> stmt s
          | Return _ :: _ -> raise (Failure ("In " ^ func.f_name ^ ", illegal" ^
              " code following return"))
          | Block sl :: ss -> check_block (sl @ ss)
          | s :: ss -> stmt s ; check_block ss
          | [] -> ()
            in check_block sl
      | Expr e -> ignore (expr e)
      | For(e1, e2, e3, st) -> ignore (expr e1); check_bool_expr e2;
        ignore (expr e3); stmt st
      | ForEach(e1, e2, st) -> let ltyp = expr e1 and rtyp = expr e2 in
        (match (ltyp, rtyp) with
            (t1, ListTyp(t2)) ->
            if t1 <> t2 then
              raise (Failure ("In " ^ func.f_name ^ ", type mismatch " ^
              string_of_typ t1 ^ " != " ^ string_of_typ t2 ^ " in " ^
              string_of_expr e1 ^ " in " ^ string_of_expr e2))
            else
              stmt st
          | (_, _) -> raise (Failure ("In " ^ func.f_name ^ ", expected " ^
              "identifier and list in, " ^ string_of_expr e1 ^ " in " ^
              string_of_expr e2)))
      | If(p, b1, b2) -> check_bool_expr p; stmt b1; stmt b2
      | Pass(_, _) -> raise (Failure ("In " ^ func.f_name ^ ", illegal use of" ^
          " pass"))
      | Return e -> let t = expr e in
        if t = func.f_typ
          then ()
        else
          raise (Failure ("In " ^ func.f_name ^ ", return type is " ^
          string_of_typ t ^ " but expecting " ^
          string_of_typ func.f_typ ^ " in " ^ string_of_expr e))
      | While(p, s) -> check_bool_expr p; stmt s

    (* Call stmt on the func body *)
    in
    stmt (Block func.f_body)

  (* Call check_function on each function in the function list *)
  in
  List.iter check_function functions;


(* Checking Nodes *************************************************************)

  (* Check that there are no duplicate node types *)
  report_duplicate (fun n -> "Duplicate node " ^ n) (List.map (fun node ->
    node.n_typ) nodes);

  (* A function that is used to check each node (node) in the nodes
     node_decl list *)
  let check_node node =

    (* Checking data block for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_data;

    (* Checking edge block for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_edge;

    (* Checking pack block for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_pack;

    (* Checking do block formals for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_do.f_formals;

    (* Checking do block locals for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_do.f_locals;

    (* Checking catch block for undefined references to tuples or nodes *)
    List.iter (fun (typ, _) -> check_defined_nt typ (node.n_typ ^ " do block");
    ()) node.n_catch.f_locals;

    (* Checks data block for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in data block")) node.n_data;

    (* Checks edge block for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in edge block")) node.n_edge;

    (* Checks pack block for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in pack block")) node.n_pack;

    (* Checks do block formals for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in do formals")) node.n_do.f_formals;

    (* Checks do block locals for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in do locals")) node.n_do.f_locals;

    (* Checks catch block locals for void types *)
    List.iter (check_not_void (fun n -> "In " ^ node.n_typ ^ ", illegal void " ^
      "type " ^ n ^ " in catch locals")) node.n_catch.f_locals;

    (* Checks data block for pack types *)
    List.iter (check_not_packet (fun n -> "In " ^ node.n_typ ^ ", illegal " ^
      "pack type " ^ n ^ " in data block")) node.n_data;

    (* Checks edge block for pack types *)
    List.iter (check_not_packet (fun n -> "In " ^ node.n_typ ^ ", illegal " ^
      "pack type " ^ n ^ " in edge block")) node.n_edge;

    (* Checks pack block for pack types *)
    List.iter (check_not_packet (fun n -> "In " ^ node.n_typ ^ ", illegal " ^
      "pack type " ^ n ^ " in pack block")) node.n_pack;

    (* Checks do block formals for pack types *)
    List.iter (check_not_packet (fun n -> "In " ^ node.n_typ ^ ", illegal " ^
      "pack type " ^ n ^ " in do formals")) node.n_do.f_formals;

    (* Checks data block for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in data block") (List.map snd node.n_data);

    (* Checks edge block for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in edge block") (List.map snd node.n_edge);

    (* Checks pack block for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in pack block") (List.map snd node.n_pack);

    (* Checks do block formals for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in do formals") (List.map snd node.n_do.f_formals);

    (* Checks do block locals for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in do locals") (List.map snd node.n_do.f_locals);

    (* Checks catch block locals for duplicates *)
    report_duplicate (fun n -> "In " ^ node.n_typ ^ ", duplicate variable " ^
      n ^ " in catch locals") (List.map snd node.n_catch.f_locals);

  (* Call check_node on each node in the node list *)
  in
  List.iter check_node nodes;

  (* A function that is used to check the do block in each node in the nodes
     list *) 
  let check_node_do_block node =

    (* Builds a string map (s_name --> s_typ) *)
    let symbols = List.fold_left (fun strmap (var_type, var_name) ->
      let res_var_type = match var_type with
        Packet ->
          let node_typ = StringMap.find "self" strmap in
          (match node_typ with
            NodeTyp n_str -> PackTyp (n_str)
          | _ -> raise (Failure ("pack type not found")))
      | typ -> typ
      in   
      StringMap.add var_name res_var_type strmap) StringMap.empty (globals @
        node.n_do.f_formals @ node.n_do.f_locals)
    in

    (* A function that takes a s_name, looks up the s_name in the
       s_name --> s_typ string map, and returns the s_typ, or an exception
       if the s_name is not found in the string map *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " do block, " ^
      "undeclared identifier " ^ s))
    in

    (* A function that returns the type of the member *)
    let type_of_member e s =
      let e_typ = e in
      let attr_dict = try TypMap.find e_typ typedefs 
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " do block, " ^
      string_of_typ e_typ ^ " type not found")) in
      try StringMap.find s attr_dict 
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " do block, " ^
      s ^ " member not found"))
    in     

    (* A recurive function that validates returns the type of an expression, or
       throws an exception if the expression is invalid *)
    let rec expr = function
        Assign(e1, e2) as exp -> let lt = 
          expr e1 and rt = expr e2 in
            check_assign lt rt (Failure ("In " ^ node.n_typ ^ " do block, " ^ 
            "illegal assignment " ^ string_of_typ lt ^ " = " ^ string_of_typ
            rt ^ " in " ^ string_of_expr exp))
      | Binop(e1, op, e2) as e -> let t1 = expr e1 and t2 = expr e2 in
        (match op with
            Add       when t1 = Float && t2 = Float -> Float
          | Add       when t1 = Int && t2 = Int -> Int
          | And       when t1 = Bool && t2 = Bool -> Bool
          | Div       when t1 = Float && t2 = Float -> Float
          | Div       when t1 = Int && t2 = Int -> Int
          | Equal     when t1 = t2 -> Bool
          | Geq       when t1 = Float && t2 = Float -> Bool
          | Geq       when t1 = Int && t2 = Int -> Bool
          | Greater   when t1 = Float && t2 = Float -> Bool
          | Greater   when t1 = Int && t2 = Int -> Bool
          | Leq       when t1 = Float && t2 = Float -> Bool
          | Leq       when t1 = Int && t2 = Int -> Bool
          | Less      when t1 = Float && t2 = Float -> Bool
          | Less      when t1 = Int && t2 = Int -> Bool
          | Mod       when t1 = Int && t2 = Int -> Int
          | Mult      when t1 = Float && t2 = Float -> Float
          | Mult      when t1 = Int && t2 = Int -> Int
          | NEqual    when t1 = t2 -> Bool
          | Or        when t1 = Bool && t2 = Bool -> Bool
          | Sub       when t1 = Float && t2 = Float -> Float
          | Sub       when t1 = Int && t2 = Int -> Int
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, illegal " ^
            "binary operator " ^ string_of_typ t1 ^ " " ^ string_of_op op ^ " "
            ^ string_of_typ t2 ^ " in " ^ string_of_expr e)))
      | BoolLit(_) -> Bool
      | Call(s, el) as call -> let func_decl = function_decl s in
        if List.length el <> List.length func_decl.f_formals then
          raise (Failure ("In " ^ node.n_typ ^ "do block, expecting " ^
          string_of_int (List.length func_decl.f_formals) ^ " arguments in " ^
          "call to " ^ string_of_expr call))
        else
          List.iter2 (fun (formal, _) exp -> let actual = expr exp in
            ignore (check_assign formal actual (Failure ("In " ^ node.n_typ ^
            " do block, illegal actual argument " ^ string_of_typ actual ^
            ", expected " ^ string_of_typ formal ^ " in parameter " ^
            string_of_expr exp ^ " in call to " ^ string_of_expr call))))
            func_decl.f_formals el; func_decl.f_typ
      | FloatLit(_) -> Float
      | Graph(_, _) -> Void
      | Id(s) -> type_of_identifier s
      | IntLit(_) -> Int
      | Lst (t, _) -> ListTyp(t)
      | ListAdd(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in 
              if t <> ele then
                raise (Failure ("In " ^ node.n_typ ^ " do block, illegal list "
                ^ "addition, " ^ string_of_typ lst ^ " += " ^ string_of_typ ele
                ^ " in " ^ string_of_expr e1 ^ " += " ^ string_of_expr e2))
              else
                ListTyp(t)
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, expected left"
            ^ " side of operator to be of type list, in " ^ string_of_expr e1 ^
            " += " ^ string_of_expr e2)))
      | ListRmv(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in (match t with
                NodeTyp(_) | TupleTyp(_) ->
                  if t <> ele then
                    raise (Failure ("In " ^ node.n_typ ^ " do block, type " ^
                    "mismatch in list removal, " ^ string_of_typ lst ^ " -= " ^
                    string_of_typ ele ^ " in " ^ string_of_expr e1 ^ " -= " ^
                    string_of_expr e2))
                  else
                    ListTyp(t)
              | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, illegal" ^
                " list removal of non-mutable list type, " ^ string_of_typ lst ^
                " -= " ^ string_of_typ ele ^ " in " ^ string_of_expr e1 ^ " -= "
                ^ string_of_expr e2)))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, expected left"
            ^ " side of operator to be of type list, in " ^ string_of_expr e1 ^
            " -= " ^ string_of_expr e2))) 
      | Member(e, s) -> type_of_member (expr e) s
      | Message -> raise (Failure ("In " ^ node.n_typ ^ " do block, illegal " ^
          "use of msg"))
      | Node(s, _) -> NodeTyp(s)
      | Noexpr -> Void
      | Null(t) -> t
      | Pack(_) -> PackTyp(node.n_typ)
      | Print(_, _) -> Void
      | Run(n, _) -> let n_decl = try (List.find (fun node_decl ->
        node_decl.n_typ = string_of_typ (expr n)) nodes) with Not_found ->
        raise (Failure ("In " ^ node.n_typ ^ " do block, did not find " ^
        string_of_expr n)) in n_decl.n_do.f_typ
      | Self -> expr (Id("self"))
      | StrLit(_) -> Str
      | Subscript(e1, e2) -> let ltyp = expr e1 in
        (match ltyp with
            ListTyp(t) -> let rtyp = expr e2 in
              (match rtyp with
                  Int -> t
                | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, " ^
                  " expected integer index type for list subscript access, in "
                  ^ string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ]")))
          | NodeTyp(t1) -> let rtyp = expr e2 in
              (match rtyp with
                  NodeTyp(t2) when t1 = t2 -> EdgeTyp(t2)
                | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, " ^
                  " expected node index type for node subscript access, in " ^
                  string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ]")))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, expected left"
            ^ " side of operator to be of type list or node, in " ^
            string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ] ")))
      | Tuple(s, _) -> TupleTyp(s)
      | Unop(uop, e) as ex -> let t = expr e in
        (match uop with
            Neg when t = Int -> Int
          | Not -> (match t with
          	| Bool -> Bool
    		| NodeTyp _ -> Bool 
    		| _ -> raise (Failure ("In " ^ node.n_typ ^ "do block, illegal unary " ^
                  "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
                      string_of_expr ex)))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " do block, illegal unary " ^
            "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
            string_of_expr ex)))
    in

    (* A function that checks if the expression is a boolean type, and throws
       an exception if it isn't *)
    let check_bool_expr e =
      if expr e <> Bool then
        raise (Failure ("In " ^ node.n_typ ^ " do block, expected a boolean " ^
        "expression, in " ^ string_of_expr e))
      else ()
    in

    (* A recursive function that validates a statement, and throws an
       exception if the statment is invalid *)
    let rec stmt = function
        Block sl -> let rec check_block = function
            [Return _ as s ] -> stmt s
          | Return _ :: _ -> raise (Failure ("In " ^ node.n_typ ^ " do block," ^
              " code following return"))
          | Block sl :: ss -> check_block (sl @ ss)
          | s :: ss -> stmt s ; check_block ss
          | [] -> ()
            in check_block sl
      | Expr e -> ignore (expr e)
      | For(e1, e2, e3, st) -> ignore (expr e1); check_bool_expr e2;
        ignore (expr e3); stmt st
      | ForEach(e1, e2, st) -> let ltyp = expr e1 and rtyp = expr e2 in
        (match (ltyp, rtyp) with
            (t1, ListTyp(t2)) ->
            if t1 <> t2 then
              raise (Failure ("In " ^ node.n_typ ^ " do block, type mismatch " ^
              string_of_typ t1 ^ " != " ^ string_of_typ t2 ^ " in " ^
              string_of_expr e1 ^ " in " ^ string_of_expr e2))
            else
              stmt st
          | (_, _) -> raise (Failure ("In " ^ node.n_typ ^ " do block, expected"
              ^ " identifier and list in, " ^ string_of_expr e1 ^ " in " ^
              string_of_expr e2)))
      | If(p, b1, b2) -> check_bool_expr p; stmt b1; stmt b2
      | Pass(_, _) -> ()
      | Return e -> let t = expr e in
        if t = node.n_do.f_typ
          then ()
        else
          raise (Failure ("In " ^ node.n_typ ^ " do block, return type is " ^
          string_of_typ t ^ " but expecting " ^ node.n_typ ^ " in " ^ 
          string_of_expr e))
      | While(p, s) -> check_bool_expr p; stmt s

    (* Call stmt on the n_do body *)
    in
    stmt (Block node.n_do.f_body)

  (* Call check_node_do_block on each node in the node list *)
  in
  List.iter check_node_do_block nodes;

  (* A function that is used to check the catch block in each node in the nodes
     list *) 
  let check_node_catch_block node =

    (* Builds a string map (s_name --> s_typ) *)
    let symbols = List.fold_left (fun strmap (var_type, var_name) ->
      let res_var_type = match var_type with
        Packet ->
          let node_typ = StringMap.find "self" strmap in
          (match node_typ with
            NodeTyp n_str -> PackTyp (n_str)
          | _ -> raise (Failure ("pack type not found")))
      | typ -> typ
      in   
      StringMap.add var_name res_var_type strmap) StringMap.empty (globals @
        node.n_catch.f_formals @ node.n_catch.f_locals)
    in

    (* A function that takes a s_name, looks up the s_name in the
       s_name --> s_typ string map, and returns the s_typ, or an exception
       if the s_name is not found in the string map *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
      "undeclared identifier " ^ s))
    in

    (* A function that returns the type of the member *)
    let type_of_member e s =
      let e_typ = e in
      let attr_dict = try TypMap.find e_typ typedefs 
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
      string_of_typ e_typ ^ " type not found")) in
      try StringMap.find s attr_dict 
      with Not_found -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
      s ^ " member not found"))
    in

    (* A recurive function that validates returns the type of an expression, or
       throws an exception if the expression is invalid *)
    let rec expr = function
        Assign(e1, e2) as exp -> let lt = 
          expr e1 and rt = expr e2 in
            check_assign lt rt (Failure ("In " ^ node.n_typ ^ " catch block, " ^ 
            "illegal assignment " ^ string_of_typ lt ^ " = " ^ string_of_typ
            rt ^ " in " ^ string_of_expr exp))
      | Binop(e1, op, e2) as e -> let t1 = expr e1 and t2 = expr e2 in
        (match op with
            Add       when t1 = Float && t2 = Float -> Float
          | Add       when t1 = Int && t2 = Int -> Int
          | And       when t1 = Bool && t2 = Bool -> Bool
          | Div       when t1 = Float && t2 = Float -> Float
          | Div       when t1 = Int && t2 = Int -> Int
          | Equal     when t1 = t2 -> Bool
          | Geq       when t1 = Float && t2 = Float -> Bool
          | Geq       when t1 = Int && t2 = Int -> Bool
          | Greater   when t1 = Float && t2 = Float -> Bool
          | Greater   when t1 = Int && t2 = Int -> Bool
          | Leq       when t1 = Float && t2 = Float -> Bool
          | Leq       when t1 = Int && t2 = Int -> Bool
          | Less      when t1 = Float && t2 = Float -> Bool
          | Less      when t1 = Int && t2 = Int -> Bool
          | Mod       when t1 = Int && t2 = Int -> Int
          | Mult      when t1 = Float && t2 = Float -> Float
          | Mult      when t1 = Int && t2 = Int -> Int
          | NEqual    when t1 = t2 -> Bool
          | Or        when t1 = Bool && t2 = Bool -> Bool
          | Sub       when t1 = Float && t2 = Float -> Float
          | Sub       when t1 = Int && t2 = Int -> Int
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, illegal " ^
            "binary operator " ^ string_of_typ t1 ^ " " ^ string_of_op op ^ " "
            ^ string_of_typ t2 ^ " in " ^ string_of_expr e)))
      | BoolLit(_) -> Bool
      | Call(s, el) as call -> let func_decl = function_decl s in
        if List.length el <> List.length func_decl.f_formals then
          raise (Failure ("In " ^ node.n_typ ^ "catch block, expecting " ^
          string_of_int (List.length func_decl.f_formals) ^ " arguments in " ^
          "call to " ^ string_of_expr call))
        else
          List.iter2 (fun (formal, _) exp -> let actual = expr exp in
            ignore (check_assign formal actual (Failure ("In " ^ node.n_typ ^
            " catch block, illegal actual argument " ^ string_of_typ actual ^
            ", expected " ^ string_of_typ formal ^ " in parameter " ^
            string_of_expr exp ^ " in call to " ^ string_of_expr call))))
            func_decl.f_formals el; func_decl.f_typ
      | FloatLit(_) -> Float
      | Graph(_, _) -> Void
      | Id(s) -> type_of_identifier s
      | IntLit(_) -> Int
      | Lst (t, _) -> ListTyp(t)
      | ListAdd(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in 
              if t <> ele then
                raise (Failure ("In " ^ node.n_typ ^ " catch block, illegal " ^
                "list addition, " ^ string_of_typ lst ^ " += " ^ string_of_typ
                ele ^ " in " ^ string_of_expr e1 ^ " += " ^ string_of_expr e2))
              else
                ListTyp(t)
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, expected" ^
            " left side of operator to be of type list, in " ^ string_of_expr e1
            ^ " += " ^ string_of_expr e2)))
      | ListRmv(e1, e2) -> let lst = expr e1 in
        (match lst with
            ListTyp(t) -> let ele = expr e2 in (match t with
                NodeTyp(_) | TupleTyp(_) ->
                  if t <> ele then
                    raise (Failure ("In " ^ node.n_typ ^ " catch block, type " ^
                    "mismatch in list removal, " ^ string_of_typ lst ^ " -= " ^
                    string_of_typ ele ^ " in " ^ string_of_expr e1 ^ " -= " ^
                    string_of_expr e2))
                  else
                    ListTyp(t)
              | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
                "illegal list removal of non-mutable list type, " ^
                string_of_typ lst ^ " -= " ^ string_of_typ ele ^ " in " ^
                string_of_expr e1 ^ " -= " ^ string_of_expr e2)))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, expected "
            ^ "left side of operator to be of type list, in " ^ string_of_expr
            e1 ^ " -= " ^ string_of_expr e2))) 
      | Member(e, s) -> type_of_member (expr e) s
      | Message -> PackTyp(node.n_typ) 
      | Node(s, _) -> NodeTyp(s)
      | Noexpr -> Void
      | Null(t) -> t
      | Pack(_) -> PackTyp(node.n_typ) 
      | Print(_, _) -> Void
      | Run(n, _) -> let n_decl = try (List.find (fun node_decl ->
        node_decl.n_typ = string_of_typ (expr n)) nodes) with Not_found ->
        raise (Failure ("In " ^ node.n_typ ^ " do catch, did not find " ^
        string_of_expr n)) in n_decl.n_do.f_typ
      | Self -> expr (Id("self"))
      | StrLit(_) -> Str
      | Subscript(e1, e2) -> let ltyp = expr e1 in
        (match ltyp with
            ListTyp(t) -> let rtyp = expr e2 in
              (match rtyp with
                  Int -> t
                | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
                  " expected integer index type for list subscript access, in "
                  ^ string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ]")))
          | NodeTyp(t1) -> let rtyp = expr e2 in
              (match rtyp with
                  NodeTyp(t2) when t1 = t2 -> EdgeTyp(t2)
                | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
                  " expected node index type for node subscript access, in " ^
                  string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ]")))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, expected" ^
            " left side of operator to be of type list or node, in " ^
            string_of_expr e1 ^ "[ " ^ string_of_expr e2 ^ " ] ")))
      | Tuple(s, _) -> TupleTyp(s)
      | Unop(uop, e) as ex -> let t = expr e in
        (match uop with
            Neg when t = Int -> Int
          | Not -> (match t with
          	| Bool -> Bool
    		| NodeTyp _ -> Bool 
    		| _ -> raise (Failure ("In " ^ node.n_typ ^ "catch block, illegal unary " ^
                  "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
                      string_of_expr ex)))
          | _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block, illegal unary " ^
            "operator " ^ string_of_typ t ^ " " ^ string_of_uop uop ^ " in " ^
            string_of_expr ex)))
    in

    (* A function that checks if the expression is a boolean type, and throws
       an exception if it isn't *)
    let check_bool_expr e =
      if expr e <> Bool then
        raise (Failure ("In " ^ node.n_typ ^ " catch block, expected a boolean "
        ^ "expression, in " ^ string_of_expr e))
      else ()
    in

    (* A recursive function that validates a statement, and throws an
       exception if the statment is invalid *)
    let rec stmt = function
        Block sl -> let rec check_block = function
            [Return _ as s ] -> stmt s
          | Return _ :: _ -> raise (Failure ("In " ^ node.n_typ ^ " catch block"
              ^ ", code following return"))
          | Block sl :: ss -> check_block (sl @ ss)
          | s :: ss -> stmt s ; check_block ss
          | [] -> ()
            in check_block sl
      | Expr e -> ignore (expr e)
      | For(e1, e2, e3, st) -> ignore (expr e1); check_bool_expr e2;
        ignore (expr e3); stmt st
      | ForEach(e1, e2, st) -> let ltyp = expr e1 and rtyp = expr e2 in
        (match (ltyp, rtyp) with
            (t1, ListTyp(t2)) ->
            if t1 <> t2 then
              raise (Failure ("In " ^ node.n_typ ^ " catch block, type " ^
              "mismatch" ^ string_of_typ t1 ^ " != " ^ string_of_typ t2 ^ " in "
              ^ string_of_expr e1 ^ " in " ^ string_of_expr e2))
            else
              stmt st
          | (_, _) -> raise (Failure ("In " ^ node.n_typ ^ " catch block, " ^
              "expected identifier and list in, " ^ string_of_expr e1 ^ " in " ^
              string_of_expr e2)))
      | If(p, b1, b2) -> check_bool_expr p; stmt b1; stmt b2
      | Pass(_, _) -> ()
      | Return e -> let t = expr e in
        if t = Void
          then ()
        else
          raise (Failure ("In " ^ node.n_typ ^ " catch block, return type is " ^
          string_of_typ t ^ " but expecting " ^ node.n_typ ^ " in " ^ 
          string_of_expr e))
      | While(p, s) -> check_bool_expr p; stmt s

    (* Call stmt on the n_catch body *)
    in
    stmt (Block node.n_catch.f_body)

  (* Call check_node_catch_block on each node in the node list *)
  in
  List.iter check_node_catch_block nodes;
